import math, Image, ImageDraw
import time

class ctm:
    "primitive: current tranforation matrix"

    # CTM_TYPE - type object
    # revamp with array module?
    def __init__(self, m11=1, m12=0, m21=0, m22=1, t1=0, t2=0):
        # default is the identity matrix
        self.mat = [m11, m12, 0, m21, m22, 0, t1, t2, 1]
    
    def dispMat(self):
        print self.mat[:3]
        print self.mat[3:6]
        print self.mat[6:9]

    def setMat(self, m11, m12, m21, m22, t1, t2):
        self.mat = [m11, m12, 0, m21, m22, 0, t1, t2, 1]
        print "matrix set to", self.mat

    def offset(self):
        "return translation coefficients as a point object"
        return self.mat[6:8]

# consts
I = ctm()
MAXMAT = 64
MAXPOINTS = 512
ONEHALF = 1<<15
MAXPOINTS = 256
ONEHALF = 1<<15

class Draw:
    "a drawing module. if no image passed in, creates empty 1000x1000 px image. enjoy."

    def __init__(self, im=None):
        self.gmats = []
        self.cp = 0   # pointer for gmats stack
        self.gctm = ctm()
        self.cexcess = 0
        self.pen = ()
        self.path = []
        self.bpoints = []
        self.size = (1000,1000)
        self.aa = 2.5
        if im is None:
            self.im = Image.new("RGB", self.size,(255,255,255)) 
        else:
            self.im = im
        self.d = ImageDraw.Draw(self.im)
        self.myStroke = -1
        self.myFill = -1

    # utils
    def _bezier(self, px0, py0, px1, py1, px2, py2, px3, py3, index):
        "draws a bezier curve using the DeCastljau algorithm"
        ptnum = [0]

        cpx0 = self._round(px0)<<16
        cpy0 = self._round(py0)<<16
        cpx1 = self._round(px1)<<16
        cpy1 = self._round(py1)<<16
        cpx2 = self._round(px2)<<16
        cpy2 = self._round(py2)<<16
        cpx3 = self._round(px3)<<16
        cpy3 = self._round(py3)<<16
        
        self._bezierh(cpx0, cpy0, cpx1, cpy1, cpx2, cpy2, cpx3, cpy3, ptnum)
        self.bpoints[ptnum[0]:] = [(px3, py3)]
        self.path[index:index] = self.bpoints
        

    def _round(self, x):
        if x-int(x) >= 0.5: return int(x+1)
        else: return int(x)
        
    def _bezierh(self, x0, y0, x1, y1, x2, y2, x3, y3, ptr):
        x01 = (x0+x1)>>1
        x12 = (x1+x2)>>1
        x23 = (x2+x3)>>1
        x02 = (x01+x12)>>1
        x13 = (x12+x23)>>1
        x03 = (x02+x13)>>1
        y01 = (y0+y1)>>1
        y12 = (y1+y2)>>1
        y23 = (y2+y3)>>1
        y02 = (y01+y12)>>1
        y13 = (y12+y23)>>1
        y03 = (y02+y13)>>1
        if (x0<x03): dx = (x03-x0)>>16
        else: dx = (x0-x03)>>16
        if (y0<y03): dy = (y03-y0)>>16
        else: dy = (y0-y03)>>16
        if (dx+dy>10):   # control resolution
            self._bezierh(x0, y0, x01, y01, x02, y02, x03, y03, ptr)
        h = (x03 + ONEHALF) >> 16
        v = (y03 + ONEHALF) >> 16
        self.bpoints[ptr[0]:] = [(h, v)]
        ptr[0] = ptr[0] + 1
            
        if (x0<x03): dx = (x03-x0)>>16
        else: dx = (x0-x03)>>16
        if (y0<y03): dy = (y03-y0)>>16
        else: dy = (y0-y03)>>16

        if (dx+dy>10):  # control resolution
            self._bezierh(x03, y03, x13, y13, x23, y23, x3, y3, ptr)    
    def antialias(self, aa=2.5):   # modify?, not used right now
        "antialias - enter factor, default 2.5"
        self.aa = aa
        
    def compMat(self, a, b):
        c_m11 = a.mat[0]*b.mat[0]+a.mat[1]*b.mat[3]+a.mat[2]*b.mat[6]
        c_m12 = a.mat[0]*b.mat[1]+a.mat[1]*b.mat[4]+a.mat[2]*b.mat[7]

        c_m21 = a.mat[3]*b.mat[0]+a.mat[4]*b.mat[3]+a.mat[5]*b.mat[6]
        c_m22 = a.mat[3]*b.mat[1]+a.mat[4]*b.mat[4]+a.mat[5]*b.mat[7]
        
        c_t1 = a.mat[6]*b.mat[0]+a.mat[7]*b.mat[3]+a.mat[8]*b.mat[6]
        c_t2 = a.mat[6]*b.mat[1]+a.mat[7]*b.mat[4]+a.mat[8]*b.mat[7]

        return ctm(c_m11, c_m12, c_m21, c_m22, c_t1, c_t2)

    def compPenX(self, x, y, a):
        b = [x, y, 1]
        return a.mat[0]*b[0]+a.mat[3]*b[1]+a.mat[6]*b[2]

    def compPenY(self, x, y, a):
        b = [x, y, 1]
        return a.mat[1]*b[0]+a.mat[4]*b[1]+a.mat[7]*b[2]
    
    def transMat(self, dx, dy):
        return ctm(1, 0, 0, 1, dx, dy)

    def transReset(self, a):   # erase?
        mattemp = a
        mattemp.mat[6] = 0
        mattemp.mat[7] = 0
        return mattemp
    
    def scaleMat(self, sx, sy):
        return ctm(sx, 0, 0, sy, 0, 0)

    def scaleReset(self, a):   # erase?
        mattemp = a
        mattemp.mat[0] = 1
        mattemp.mat[4] = 1
        return mattemp

    def skewXMat(self, rtheta):   # angle in degrees
        theta = rtheta*pi/180
        return ctm(1, 0, math.tan(theta), 1, 0, 0)

    def skewYMat(self, rtheta):   # angle in degrees
        theta = rtheta*pi/180
        return ctm(1, math.tan(theta), 0, 1, 0, 0)

    def rotMat(self, rtheta):
        theta = rtheta*math.pi/180
        return ctm(math.cos(theta), math.sin(theta), math.sin(theta)*-1, math.cos(theta), 0, 0)
        
    # operations    
    def copyMat(self, a, dest):
            dest.mat = a.mat

    def gmatsFull(self):
        return self.cp>=MAXMAT

    def push(self, obj=None):
        self.cexcess = self.cexcess + 1
        self.cp
        if obj is None:
            obj = self.gctm
        if self.gmatsFull():
            print "CTM overflow!", self.cexcess   # add %d 
        self.gmats[self.cp:] = [self.gctm]   # : => insert/replace
        self.cp = self.cp + 1
        self.gctm = obj
        
        for x in self.gmats:
            if x != 0:
                print "   gmat ", x.mat
        print "   cp", self.cp
    
    def pop(self):   
        self.cexcess = self.cexcess - 1
        self.cp = self.cp - 1
        if self.cp<0:
            print "CTM underflow!"
        self.gctm = self.gmats[self.cp]
        
##        for x in self.gmats:
##            if x != 0:
##                print "   gmat ", x.mat
##        print "   cp", self.cp
        
        self.gmats[self.cp:] = []   # clean up stack
        self.myStroke = 0
        self.myFill = 0
        
    def rotate(self, t):   # counterclockwise
        mattemp = self.rotMat(t)
        self.gctm = self.compMat(mattemp, self.gctm)
       
    def scale(self, x, y):
        mattemp = self.scaleMat(x, y)
        self.gctm = self.compMat(mattemp, self.gctm)

    def show(self):
        self.im.show()
        
    def skewx(self, t):
        mattemp = self.skewMat(x, y)
        self.gctm = self.compMat(mattemp, self.gctm)
    
    def skewy(self, t):
        mattemp = self.skewYMat(t)
        self.gctm = self.compMat(mattemp, self.gctm)
  
    def translate(self, x, y):
        mattemp = self.transMat(x, y)
        self.gctm = self.compMat(mattemp, self.gctm)
 
    def moveto(self, x, y):
        #mattemp = self.transReset(self.gctm)
        penX = self.compPenX(x, y, self.gctm)
        penY = self.compPenY(x, y, self.gctm)
        self.path = []
        self.pen = (penX, penY)
        self.path.append(self.pen)
            
    def rmoveto(self, x, y):    # acts like newpath
        if self.pen == ():
            print "no point of reference!"
        else:
            penX = self.compPenX(x, y, self.gctm)
            penY = self.compPenY(x, y, self.gctm)
            self.path = []   # cleans up path
            self.pen = (penX, penY)
            self.path.append(self.pen)
       
    def lineto(self, x, y):
        if self.path is []:
            print "no current point!"
        else:
            #mattemp = self.transReset(self.gctm)
            penX = self.compPenX(x, y, self.gctm)
            penY = self.compPenY(x, y, self.gctm)
            self.pen = (penX, penY)
            self.path.append(self.pen)
        
    def rlineto(self, x, y):   
        if self.path is []:
            print "no current point!"
        else:
            penX = self.compPenX(x, y, self.gctm)
            penY = self.compPenY(x, y, self.gctm)
            self.pen = (penX, penY)
            self.path.append(self.pen)

    def curveto(self, x1, y1, x2, y2, x3, y3):
        if self.path is []:
            print "no current point!"
        else:
            #mattemp = self.transReset(self.gctm)
            penX1 = self.compPenX(x1, y1, self.gctm)
            penY1 = self.compPenY(x1, y1, self.gctm)
            penX2 = self.compPenX(x2, y2, self.gctm)
            penY2 = self.compPenY(x2, y2, self.gctm)
            penX3 = self.compPenX(x3, y3, self.gctm)
            penY3 = self.compPenY(x3, y3, self.gctm)
            self.pen = (penX3, penY3)
            self.path.append((penX1,penY1, penX2,penY2, penX3,penY3))
    def rcurveto(self, x1, y1, x2, y2, x3, y3):
        if self.path is []:
            print "no current point!"
        else:
            penX1 = self.compPenX(x1, y1, self.gctm)
            penY1 = self.compPenY(x1, y1, self.gctm)
            penX2 = self.compPenX(x2, y2, self.gctm)
            penY2 = self.compPenY(x2, y2, self.gctm)
            penX3 = self.compPenX(x3, y3, self.gctm)
            penY3 = self.compPenY(x3, y3, self.gctm)
            self.pen = (penX3, penY3)
            self.path.append((penX1,penY1, penX2,penY2, penX3,penY3))

    def closepath(self):
        firstpt = self.path[0]
        lasti = len(self.path) - 1
        lastpt = self.path[lasti]
        if lastpt != firstpt:
            self.path.append(firstpt)
           
    def stroke(self, rgb=1):
        self.myStroke = rgb
        maxPath = len(self.path)
        for i in range (0, maxPath):
            if len(self.path[i]) == 6:
                sPt = self.path[i-1]
                rPts = self.path[i]
                del self.path[i]
                self._bezier(sPt[0], sPt[1], rPts[0], rPts[1], rPts[2], rPts[3], rPts[4], rPts[5], i)
        #print "about to stroke: ", self.path
        if len(self.path) < 2:
            print "invalid stroke path!"
        else:
            #self.im.resize((self.size[0]*self.aa,self.size[1]*self.aa), Image.ANTIALIAS)
            lastLine = len(self.path)-1
            for i in range(0, lastLine):
                self.d.line([self.path[i][0], self.path[i][1], self.path[i+1][0], self.path[i+1][1]], fill=self.myStroke)
            #self.im.resize(self.size, Image.ANTIALIAS)
            
    def fill(self, fillRGB=1, strokeRGB=None):
        self.myFill = fillRGB
        self.myStroke = strokeRGB
        maxPath = len(self.path)
        for i in range (0, maxPath):
            if len(self.path[i]) == 6:
                sPt = self.path[i-1]
                rPts = self.path[i]
                del self.path[i]
                self._bezier(sPt[0], sPt[1], rPts[0], rPts[1], rPts[2], rPts[3], rPts[4], rPts[5], i)
        print "about to fill: ", self.path
        if len(self.path) < 3:
            print "invalid fill path!"
        else:
            # print "prior to resize: ", self.im.size, (self.size[0]*self.aa,self.size[1]*self.aa)
            self.im.resize((self.size[0]*self.aa,self.size[1]*self.aa), Image.ANTIALIAS)
            # print "after resize: ", self.im.size
            # print "just before polygon with color: ", self.myFill
            if self.myStroke is None:
                self.d.polygon(self.path, fill=self.myFill, outline=self.myFill)
            else:
                self.d.polygon(self.path, fill=self.myFill, outline=self.myStroke)
            # print "prior to resize: ", self.im.size
            self.im.resize(self.size, Image.ANTIALIAS)    
               
print "*******************************************"
print "MimiDraw loaded!"


# --------------------------------------------------------------
##
##def pac():
##    test.scale(100, 100)
##    test.rotate(66)
##    test.moveto(1,1)
##    test.lineto(0,0)
##    test.curveto(0,2,2,2,2,0)
##    test.fill((0,50,155))
##   
test = Draw()
t0 = time.clock()
test.push()
##test.moveto(0,0)
##test.curveto(1000,1000,1000,0,0,1000)
##test.stroke()
for i in range(0, 100):
    test.moveto(10, i*10)
    test.curveto(330, i*10+100, 990, i*10+200, 990, i*10+300)
    #test.lineto(990,i+100)
    test.rotate(30)
    test.stroke()
test.pop()
print "time for 100 skew lines:", time.clock()-t0
test.show()

##
##test.translate(700,700)
##pac()
##test.pop()
##test.push()
##test.translate(200,700)
##pac()
##test.pop()
##test.im = test.im.resize((500,500), Image.ANTIALIAS)
##test.show()
##print "size: ", test.im.size


    
